# platforms.py

def get_methods():
    """
    Return a dictionary with methods to deploy and test functions for each
    platform (cdf, gc or local). This decouples platforms and deployment
    """
    methods={}
    
    # cdf
    from akerbp.mlops.cdf.helpers import deploy_function as cdf_deploy
    from akerbp.mlops.cdf.helpers import test_function as cdf_test
    from akerbp.mlops.cdf.helpers import set_up_cdf_client
    set_up_cdf_client(context='deploy')
    methods["cdf"] = cdf_deploy, cdf_test
    
    # gc
    from akerbp.mlops.gc.helpers import deploy_function as gc_deploy
    from akerbp.mlops.gc.helpers import test_function as gc_test
    methods["gc"] = gc_deploy, gc_test
    
    # local (methods don't do anything)
    local_deploy = local_test = lambda *args,**kargs: None
    methods["local"] = local_deploy, local_test
    
    return methods


"""
service.py

Prediction service. 
"""
from importlib import import_module

from akerbp.mlops import __version__ as version
from akerbp.mlops.core import config, logger

c = config.read_config()
model_module = import_module(c.model_import_path)
predict = model_module.predict
_initialization = model_module.initialization
ModelException = model_module.ModelException

logging=logger.get_logger("mlops_services")

logging.debug(f"MLOps framework version {version}")


def initialization(secrets):
    """
    Read initialization object required by `predict`
    """
    # This check adds a little overhead to each prediction
    if 'init_object' not in globals():
        global init_object
        artifact_folder = c.artifact_folder
        init_object = _initialization(artifact_folder, secrets)
        logging.debug(f"Loaded initialization object")


def service(data, secrets):
    """
    Generate prediction for an input
    Inputs:
        data: input to the model (sent by a user through the API)
        secrets: api keys that the model may need during initialization
    Output:
        Dictionary with a status field ('ok' or 'error'). If status is 'ok', 
         there is a 'prediction' field as well. If status is 'error', there is 
         a 'message' field as well.
    """
    try:
        initialization(secrets)
        logging.debug(f"Predict {data=}")
        y = predict(data, init_object, secrets)
        logging.debug(f"Prediction is {y=}")
        return dict(status="ok", prediction=y, model_id=c.model_id)
    except ModelException as error:
        error_message = f"Could not get a prediction. Message: {error}"
        logging.error(error_message)
        return dict(status="error", message=error_message, model_id=c.model_id)



# handler.py
import traceback
from importlib import import_module
import warnings
warnings.simplefilter("ignore")

from akerbp.mlops.core import logger, config
service_name = config.envs.service_name
logging=logger.get_logger("mlops_cdf")

from akerbp.mlops.cdf.helpers import log_system_info
log_system_info()

service = import_module(f"akerbp.mlops.services.{service_name}").service


def handle(data, secrets):
   try:
      config.update_cdf_keys(secrets)
      return service(data, secrets)
   except Exception:
      trace = traceback.format_exc()
      error_message = f"{service_name} service failed.\n{trace}"
      logging.critical(error_message)
      return dict(status='error', error_message=error_message)



"""
cdf_helpers.py

High level functionality built on top of Cognite's Python SDK

Most functions require the global cdf client to be set up before using them.
"""
import time
import sys
import os 
import json
from shutil import make_archive, unpack_archive

from cognite.client.exceptions import CogniteNotFoundError, CogniteAPIError

from akerbp.mlops.core import logger, config

logging=logger.get_logger(name='mlops_cdf')

global_client = {}
api_keys=config.api_keys


def set_up_cdf_client(context='run'):
    """
    Set up the global client used by most helpers. This needs to be called
    before using any helper. 

    Input:
      - context: string either 'run' (access to data and functions) or 'deploy'
        (access to 'functions' also).
    """
    if context == 'run':
        api_key_labels=["data", "files"]
    elif context == 'deploy':
        api_key_labels=["data", "files", "functions"]
    else:
        raise ValueError("Context should be either 'run' or 'deploy'")

    for k in api_key_labels:
        if k not in global_client:
            global_client[k] = get_client(api_keys[k], k)
        
    logging.info("CDF client was set up correctly")


def get_client(api_key, api_key_label=None):
    """
    Create a CDF client with a given api key
    """
    if api_key_label == 'functions':
        from cognite.experimental import CogniteClient
        logging.warning("Imported CogniteClient from cognite.experimental")
    else:
        from cognite.client import CogniteClient
    
    client = CogniteClient(
        api_key=api_key,
        project='akbp-subsurface',
        client_name="mlops-client",
        base_url='https://api.cognitedata.com'
    )
    assert client.login.status().logged_in
    logging.debug(f"{client.version=}")
    return client


def create_function(
    function_name, 
    folder, 
    handler_path,  
    description='',
    owner='',
    secrets={}
):
    """
    Create a Cognite function from a folder. Any existing function with the same
    name is deleted first.

    Inputs:
      - function_name: name of the function to create
      - folder: path where the source code is located
      - handler_path: path to the handler file
      - secrets: api keys or similar that should be passed to the function
    """
    client = global_client["functions"]
    try:
        client.functions.delete(external_id=function_name)
        logging.debug(f"Deleted function {function_name}")
    except CogniteNotFoundError:
        pass
    env_vars = {k.upper():v for k,v in config.envs_dir.items() if v}
    env_vars.update(dict(DEPLOYMENT_PLATFORM='cdf'))
    function = client.functions.create( 
        name=function_name, 
        folder=folder, 
        function_path=handler_path,
        external_id=function_name,
        description=description,
        owner=owner,
        secrets=secrets,
        env_vars=env_vars
    )
    logging.debug(
        f"Created {function_name}: {folder=}, {handler_path=}"
    )

    return function


def deploy_function(
    function_name,
    folder='.',
    handler_path='handler.py',
    secrets=api_keys,
    info={'description':'', 'owner':''}
):
    """
    Deploys a Cognite function from a folder. The function is created first,
    and then it waits until the function status is ready or failed. If it fails,
    it will try again `max_error` times

    Inputs:
      - function_name: name of the function to create
      - folder: path where the source code is located
      - handler_path: path to the handler file
      - secrets: api keys or similar that should be passed to the function
    """
    max_errors = 3

    for trial in range(max_errors):
        function = create_function(
            function_name, 
            folder, 
            handler_path, 
            info['description'],
            info['owner'],
            secrets
        )
        status = wait_function_status(function)
        logging.debug(f"Function status is {status}")
        if function.status == 'Ready':
            break
        if function.status == 'Failed' and trial < max_errors-1:
            logging.warning(f"Function failed: {function.id=}")
            logging.debug(f"Error was: {function.error=}")
            logging.debug(f"Try to create function again")
        else:
            raise Exception(f"Function deployment error: {function.error=}")


def call_function(function_name, data):
    """
    Call a function deployed in CDF
    """
    client = global_client["functions"]
    function = client.functions.retrieve(external_id=function_name)
    logging.debug(f"Retrieved function {function_name}")
    call = function.call(data)
    logging.debug(f"Called function (call_id={call.id})")
    response = call.get_response()
    logging.debug(f"Called function: {response=}")
    return response


def test_function(function_name, data):
    """
    Call a function with data and verify that the response's 
    status is 'ok'
    """
    logging.debug(f"Test function {function_name}")
    output = call_function(function_name, data)
    assert output['status'] == 'ok'
    logging.info(f"Test call was successful :)")


def wait_function_status(function, status=["Ready", "Failed"]):
    """
    Wait until function status is in `status`
    By default it waits for Ready or Failed, which is useful when deploying.
    It implements some control logic, since polling status can fail.
    """
    polling_wait_seconds_base = 10
    polling_wait_seconds = polling_wait_seconds_base
    max_api_errors_base = 5
    max_api_errors = max_api_errors_base
    
    logging.debug(f"Wait for function to be ready or to fail")
    while not (function.status in status):
        try:
            time.sleep(polling_wait_seconds)
            function.update()
            logging.debug(f"{function.status=}")
            polling_wait_seconds = polling_wait_seconds_base
            max_api_errors = max_api_errors_base
        except CogniteAPIError as e:
            max_api_errors -= 1
            logging.warning(f"Could not update function status, will try again")
            polling_wait_seconds *= 1.2          
            if not max_api_errors:
                logging.error(f"Could not update function status.")
                raise e

    return function.status


def download_file(id, path):
    """
    Download file from Cognite
    
    Params:
        - id: dictionary with id type (either "id" or "external_id") as key
        - path: path of local file to write
    """
    client = global_client["files"]
    
    logging.debug(f"Download file with {id=} to {path}")
    client.files.download_to_path(path, **id)


def upload_file(
    external_id, 
    path, 
    metadata={}, 
    directory='/',
    overwrite=True):
    """
    Upload file to Cognite
    
    Params:
        - external_id: external id
        - path: path of local file to upload
        - metadata: dictionary with file metadata
        - overwrite: what to do when the external_id exists already
    """
    client = global_client["files"]
    
    metadata = {k:v if isinstance(v, str) else json.dumps(v) 
        for k,v in metadata.items()}

    logging.debug(f"Upload file {path} with {external_id=} and {metadata=}")
    file_info = client.files.upload(
        path, 
        external_id, 
        metadata=metadata, 
        directory=directory, 
        overwrite=overwrite
    )
    logging.info(f"Uploaded file: {file_info=}")
    return file_info


def upload_folder(
    external_id, 
    path, 
    metadata={}, 
    overwrite=False, 
    target_folder='/'
):
    """
    Upload folder content to Cognite. It compresses the folder and uploads it.
    
    Params:
        - external_id: external id (should be unique in the CDF project)
        - path: path of local folder where content is stored
        - metadata: dictionary with file metadata
        - target_folder: path where compressed file should be stored
        - overwrite: if overwrite==False and `external_id` exists => exception
    """
    base_name = os.path.join(path, 'archive')
    archive_name = make_archive(base_name, 'gztar', path)
    file_info = upload_file(
        external_id, 
        archive_name, 
        metadata=metadata,
        overwrite=overwrite,
        directory=target_folder
    )
    os.remove(archive_name)
    logging.info(f"Folder content uploaded: {file_info=}")
    return file_info


def download_folder(external_id, path):
    """
    Download content from Cognite to a folder. It is assumed to have been
    uploaded using `upload_folder()`, so it downloads a file and decompresses
    it.

    Params:
    - external_id: external id
    - path: path of local folder where content will be stored
    """
    base_name = os.path.join(path, 'archive.tar.gz')
    download_file(dict(external_id=external_id), base_name)
    unpack_archive(base_name, os.path.dirname(base_name))
    os.remove(base_name)
    logging.info(f"Model file/s downloaded to {path}")


def log_system_info():
    """
    Can be called from a handler to log CDF environment information
    """
    logging.debug(f"Python version:\n{os.popen('python --version').read()}")
    logging.debug(f"Python path:\n{sys.path}")
    logging.debug(f"Current working directory:\n{os.getcwd()}")
    logging.debug(f"Content:\n{os.popen('ls -la *').read()}")
    logging.debug(f"Packages:\n{os.popen('pip freeze').read()}")


def count_file_versions(
    external_id_prefix, 
    directory_prefix='/'
):
    """
    How many files have the given `external_id_prefix` and `directory_prefix`
    This can be used for file versioning.

    Returns an integer (>=0)
    """
    client = global_client["files"]
    result = client.files.aggregate(filter={
        "external_id_prefix": external_id_prefix, 
        "directory_prefix": directory_prefix
    })
    return result[0]["count"]


def query_file_versions(directory_prefix, external_id_prefix, metadata):
    """
    Find all file versions that match a query. Only versions that were uploaded
    are considered

    Input:
        -directory_prefix
        -external_id_prefix
        -metadata: query to the metadata (dictionary)
    Output:
        - list of versions (dataframe)
    """
    client = global_client["files"]
    file_list = client.files.list(
        limit=-1, 
        directory_prefix=directory_prefix, 
        external_id_prefix=external_id_prefix,
        metadata=metadata,
        uploaded=True
    ).to_pandas()

    return file_list



# config.py
import os
from typing import List, Optional
from pydantic.dataclasses import dataclass
from dataclasses import asdict

from akerbp.mlops.core import helpers 
from akerbp.mlops.core import logger

logging=logger.get_logger(name='mlops_core')


@dataclass
class EnvVar():
    env: str
    service_name: Optional[str]
    local_deployment: Optional[str]
    google_project_id: Optional[str]
    deployment_platform: Optional[str]

    def __post_init__(self):
        envs = ["dev", "test", "prod"]
        if self.env not in envs:
            m = (f"Environment: allowed values are {envs}"
                 f", got '{self.env}'")
            raise ValueError(m)

        service_names = ["training", "prediction"]
        if self.env != 'dev' and self.service_name not in service_names:
            m = (f"Service name: allowed values are {service_names}"
                 f", got '{self.service_name}'")
            raise ValueError(m)

        local_deployments = ["True", None]
        if self.local_deployment not in local_deployments:
            m = (f"Local deployment: allowed values are {local_deployments}"
                 f", got '{self.local_deployment}'")
            raise ValueError(m)

        deployment_platforms = ["cdf", "gc", None]
        if self.deployment_platform not in deployment_platforms:
            m = (f"Local deployment: allowed values are {deployment_platforms}"
                 f", got '{self.deployment_platforms}'")
            raise ValueError(m)

# Read environmental variables
envs=EnvVar(
    env=os.getenv('ENV'), 
    service_name=os.getenv('SERVICE_NAME'), 
    local_deployment=os.getenv('LOCAL_DEPLOYMENT'),
    google_project_id=os.getenv('GOOGLE_PROJECT_ID'),
    deployment_platform=os.getenv('DEPLOYMENT_PLATFORM'),
)
envs_dir=asdict(envs)
logging.debug(f"{envs_dir=}")

@dataclass
class CdfKeys():
    data: Optional[str]
    files: Optional[str]
    functions: Optional[str]

_api_keys = CdfKeys(
   data = os.getenv('COGNITE_API_KEY_DATA'),
   files = os.getenv('COGNITE_API_KEY_FILES'),
   functions = os.getenv('COGNITE_API_KEY_FUNCTIONS')
)
api_keys = asdict(_api_keys)

def update_cdf_keys(new_keys):
    global api_keys
    api_keys=asdict(CdfKeys(**new_keys))


@dataclass
class ServiceSettings():
    model_name: str
    model_file: str
    req_file: str
    test_file: str
    artifact_folder: str
    platform: str
    info: dict
    model_id: Optional[str] = None

    def __post_init__(self):
        # Validation
        if not os.path.isfile(self.model_file):
            raise ValueError(f"File {self.model_file} does not exist")

        if not os.path.isfile(self.req_file):
            raise ValueError(f"File {self.req_file} does not exist")

        
        if self.platform not in ["cdf", "gc", "local"]:
            m = f"Deployment platform should be either 'cdf', 'gc' or 'local'"
            raise ValueError(m)

    def __post_init_post_parse__(self):
        # Derived fields
        if envs.env == 'dev' and not envs.local_deployment:
            self.platform = 'local'
        self.model_import_path = helpers.as_import_path(self.model_file)
        self.test_import_path = helpers.as_import_path(self.test_file)

        self.files = {
            "model code": helpers.get_top_folder(self.model_file), 
            "handler": (f"akerbp.mlops.cdf","handler.py"),
            "artifact folder": self.artifact_folder
        }
        if self.platform == "gc":
            files_gc = {
                "Dockerfile": ("akerbp.mlops.gc", "Dockerfile"),
                "requirements.app": ("akerbp.mlops.gc", "requirements.app"),
                "install_req_file.sh": ("akerbp.mlops.gc", "install_req_file.sh")
            }
            self.files = {**self.files, **files_gc}


def store_service_settings(c):
    logging.info("Write service settings file")
    service_settings=asdict(c)
    with open('mlops_service_settings.py', 'w') as f:
        f.write(f'{service_settings=}')


@dataclass
class ProjectSettings():
    project_settings: List[ServiceSettings]


def read_config():
    if os.path.isfile('mlops_settings.py'):
        logging.info(f"Read project settings")
        from mlops_settings import model_names, model_files, req_files
        from mlops_settings import artifact_folders, infos, test_files
        from mlops_settings import model_platforms
        settings = [dict(
            model_name=model_names[x],
            model_file=model_files[x],
            req_file=req_files[x],
            artifact_folder=artifact_folders[x],
            info=infos[x],
            test_file=test_files[x],
            platform=model_platforms[x]
        ) for x in range(len(model_names))]
        model_settings = [ServiceSettings(**s) for s in settings]
        project_settings = ProjectSettings(project_settings=model_settings)
        logging.debug(f"{project_settings=}")        
        return project_settings
    #elif os.path.isfile('mlops_service_settings.py'):
    else:
        logging.info(f"Read service configuration")
        from mlops_service_settings import service_settings
        service_settings = ServiceSettings(**service_settings)
        logging.debug(f"{service_settings=}")
        return service_settings
    #else:
    #    raise Exception("Didn't find a suitable settings file")



"""
deploy.py

Deploy services in either Google Cloud Run or CDF Functions. 
Model registry uses CDF Files.
""" 
import os
import traceback
import shutil

import akerbp.mlops.model_manager as mm
from akerbp.mlops.deployment import helpers, platforms
from akerbp.mlops.core import logger, config
logging=logger.get_logger(name='mlops_deployment')
env = config.envs.env
service_name = config.envs.service_name



platform_methods = platforms.get_methods()


def deploy_model(model_settings, platform_methods=platform_methods):

    c = model_settings
    logging.debug(" ")
    logging.info(f"Deploy model {c.model_name}")

    deployment_folder =f'mlops_{c.model_name}'
    function_name=f"{c.model_name}-{service_name}-{env}"

    if service_name == 'prediction':
        c.model_id = mm.set_up_model_artifact(
            c.artifact_folder, 
            c.model_name
        )

    m = "Create deployment folder and move required files/folders"
    logging.info(m)
    os.mkdir(deployment_folder)
    
    helpers.copy_to_deployment_folder(c.files, deployment_folder)

    logging.debug(f"cd {deployment_folder}")
    os.chdir(deployment_folder)
    
    helpers.set_up_requirements(c)

    config.store_service_settings(c)

    # * Dependencies: (user settings <- model test). Either run before
    #   going to the dep. folder or copy project config to dep. folder. 
    # * It is important to run tests after setting up the artifact
    #   folder in case it's used to test prediction service.
    # * Tests need the model requirements installed!
    logging.info(f"Run model and service tests")
    input, check = helpers.get_model_test_data(c.test_import_path)
    if c.test_file:
        helpers.run_tests(c.test_file)
        from akerbp.mlops.services.test_service import test_service
        test_service(input, check)
    else:
        logging.warning("Model test file is missing! Didn't run tests")


    logging.info(f"Deploy {function_name} to {c.platform}")
    deploy_function, test_function = platform_methods[c.platform]
    deploy_function(function_name, info=c.info[service_name])
    
    if c.test_import_path:
        logging.info("Make a test call")
        test_function(function_name, input)
    else:
        logging.warning("No test file was set up. End-to-end test skipped!")


def deploy(project_settings):
    failed_models = {}
    base_path = os.getcwd()
        
    for c in project_settings:
        model_name = c.model_name
        deployment_folder =f'mlops_{model_name}'
        try:
            deploy_model(c)
        except Exception:
            trace = traceback.format_exc()
            error_message = f"Model failed to deploy!\n{trace}"
            logging.error(error_message)
            
            failed_models[model_name] = error_message
        finally:
            logging.debug(f"cd ..")
            os.chdir(base_path)
            logging.debug(f"Delete deployment folder")
            if os.path.isdir(deployment_folder):
                shutil.rmtree(deployment_folder)

    if failed_models:
        for model_name, error_message in failed_models.items():
            logging.debug(" ")
            logging.info(f"Model {model_name} failed: {error_message}")
        raise Exception("At least one model failed.")


if __name__ == '__main__':
    mm.setup()
    s = config.read_config().project_settings
    deploy(s)


"""
service.py

Training service
"""
from tempfile import TemporaryDirectory
from importlib import import_module

import akerbp.mlops.model_manager as mm
from akerbp.mlops import __version__ as version
from akerbp.mlops.core import config, logger 
logging=logger.get_logger("mlops_services")
c = config.read_config()
env = config.envs.env

model_module = import_module(c.model_import_path)
train = model_module.train
ModelException = model_module.ModelException

logging.debug(f"MLOps framework version {version}")


def _saver(path, metadata, cdf_api_keys):
    """
    Upload model folder to CDF Files
    """
    mm.setup(cdf_api_keys)
    model_info = mm.upload_new_model_version(
        c.model_name, 
        env,  
        path,
        metadata
    )
    return model_info


def service(data, secrets, saver=_saver):
    """
    Training service
    Inputs:
        - data: dictionary, data passed by the user through the API
        - secrets: dictionary with api keys
        - saver: an object that saves the model folder

    Output:
        - Dictionary with status 'ok' or 'error' as keys.
            status == 'ok'    -> there is a 'training' key as well  
                                (data on the model file)
            status == 'error' -> there is a 'message' key as well
    """
    try:
        with TemporaryDirectory() as temp_dir:
            metadata = train(data=data, folder_path=temp_dir, secrets=secrets)
            logging.debug(f"{metadata=}")
            model_info = saver(temp_dir, metadata, secrets)
        return dict(status="ok", training=model_info)
    except ModelException as e:
        error_message = f"Training failed. Message: {e}"
        logging.error(error_message)
        return dict(status='error', message=error_message)
        
        
    



# helpers.py
import os
from akerbp.mlops.core import logger


logging=logger.get_logger(name='mlops_core')

def get_top_folder(s):
    return s.split(os.sep)[0]


def as_import_path(file_path):
    if file_path:
        return file_path.replace(os.sep,'.').replace('.py','')
    else:
        logging.debug(f"Empty file path -> empty import path returned")



"""
setup.py 

Information used to build the package
"""
from setuptools import find_namespace_packages, setup
import os
import subprocess


with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()


def get_version():
    ENV = os.environ['ENV'] # Must be set
    tag = subprocess.check_output(
        ['git', 'describe', '--tags', '--exact-match'],
        encoding='UTF-8'
    )
    tag = tag.rstrip()
    if ENV in ['dev', 'test']:
        version = '0.' + tag
    elif ENV == 'prod':
        version = '1.' + tag
    else:
        raise ValueError(f"Unknown environment {ENV}")
    return version


setup(
    name="akerbp.mlops", 
    version=get_version(),
    author="Alfonso M. Canterla",
    author_email="alfonso.canterla@soprasteria.com",
    description="MLOps framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://bitbucket.org/akerbp/mlops/",
    packages=find_namespace_packages(where='src'),
    package_dir={'': 'src'},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "cognite-sdk-experimental>=0.34.0", 
        "pytest>=6.1.1",
        "pydantic>=1.7.3"
    ],
    scripts=[
        'src/akerbp/mlops/deployment/deploy_training_service.sh', 
        'src/akerbp/mlops/deployment/deploy_prediction_service.sh',
        'src/akerbp/mlops/gc/install_gc_sdk.sh'
    ],
    python_requires='>=3.8',
    include_package_data=True,
    package_data={'': [
        'mlops/gc/Dockerfile', 
        'mlops/gc/requirements.app',
        'mlops/gc/install_req_file.sh',
        'mlops/deployment/bitbucket-pipelines.yml' 
        ]},
)